from enum import Enum

import nonebot
import nonebot_plugin_localstore as store
from pydantic import BaseModel

data_dir = store.get_plugin_data_dir() / "word_bank"
data_dir.mkdir(parents=True, exist_ok=True)


class Config(BaseModel):

    zxwb_db_url: str | None = None
    """数据库链接，默认使用sqlite"""


ZxwbConfig = nonebot.get_plugin_config(Config)

db_type = "sqlite"

if ZxwbConfig.zxwb_db_url:
    db_type = ZxwbConfig.zxwb_db_url.split(":", 1)[0] if ZxwbConfig.zxwb_db_url else ""


class ScopeType(Enum):
    """
    全局、群聊、私聊
    """

    GLOBAL = 0
    GROUP = 1
    PRIVATE = 2


scope2int = {
    "全局": ScopeType.GLOBAL,
    "群聊": ScopeType.GROUP,
    "私聊": ScopeType.PRIVATE,
}


class WordType(Enum):
    """
    精准、模糊、正则、图片
    """

    EXACT = 0
    FUZZY = 1
    REGEX = 2
    IMAGE = 3


type2int = {
    "精准": WordType.EXACT,
    "模糊": WordType.FUZZY,
    "正则": WordType.REGEX,
    "图片": WordType.IMAGE,
}

int2type = {
    "EXACT": "精准",
    "FUZZY": "模糊",
    "REGEX": "正则",
    "IMAGE": "图片",
}



from nonebot import logger
from nonebot_plugin_uninfo import Uninfo
from zhenxun_utils.message import MessageUtils
from nonebot_plugin_alconna import (
    Args,
    Match,
    Query,
    Option,
    Alconna,
    AlconnaQuery,
    on_alconna,
    store_true,
)

from ._model import WordBank
from ._config import ScopeType
from ._data_source import WordBankManage

_show_matcher = on_alconna(
    Alconna(
        "显示词条",
        Args["problem?", str],
        Option("-g|--group", Args["gid", str], help_text="群组id"),
        Option("--id", Args["index", int], help_text="词条id"),
        Option("--all", action=store_true, help_text="全局词条"),
    ),
    aliases={"查看词条"},
    priority=5,
    block=True,
)


@_show_matcher.handle()
async def _(
    session: Uninfo,
    problem: Match[str],
    index: Match[int],
    gid: Match[str],
    all: Query[bool] = AlconnaQuery("all.value", False),
):
    group_id = session.group.id if session.group else None
    word_scope = ScopeType.GROUP if group_id else ScopeType.PRIVATE
    if all.result:
        word_scope = ScopeType.GLOBAL
    if gid.available:
        group_id = gid.result
    if problem.available:
        if index.available and (
            index.result < 0
            or index.result > len(await WordBank.get_problem_by_scope(word_scope))
        ):
            await MessageUtils.build_message("id必须在范围内...").finish(reply_to=True)
        result = await WordBankManage.show_word(
            problem.result,
            index.result if index.available else None,
            group_id,
            word_scope,
        )
    else:
        result = await WordBankManage.show_word(
            None, index.result if index.available else None, group_id, word_scope
        )
    await result.send()
    logger.info(
        f"查看词条回答: {problem.result if problem.available else index.result}"
    )



from nonebot import require
from nonebot.plugin import PluginMetadata, inherit_supported_adapters

from ._config import Config

require("nonebot_plugin_alconna")
require("nonebot_plugin_uninfo")

__plugin_meta__ = PluginMetadata(
    name="ZXWB词库问答",
    description="小真寻的词库问答，对群组/私聊/全局进行区分，支持 精准/模糊/正则/图片 匹配",
    usage=r"""
    对指定问题的随机回答，对相同问题可以设置多个不同回答
    删除词条后每个词条的id可能会变化，请查看后再删除
    更推荐使用id方式删除
    问题回答支持的类型：at, image
    查看词条命令：群聊时为 群词条+全局词条，私聊时为 私聊词条+全局词条
    添加词条正则：添加词条(模糊|正则|图片)?问\s*?(\S*\s?\S*)\s*?答\s?(\S*)
    正则问可以通过$1类推()捕获的组
    注意：可以通过引用来提供回答， 如：（引用）添加词条问你好
    指令：
        添加词条 ?[模糊|正则|图片]问...答...：添加问答词条，可重复添加相同问题的不同回答
            示例:
                添加词条问你好答你也好
                添加词条图片问答看看涩图
        删除词条 ?[问题] ?[序号] ?[回答序号]：删除指定词条指定或全部回答
            示例:
                删除词条 谁是萝莉           : 删除文字是 谁是萝莉 的词条
                删除词条 --id 2            : 删除序号为2的词条
                删除词条 谁是萝莉 --aid 2   : 删除 谁是萝莉 词条的第2个回答
                删除词条 --id 2 --aid 2    : 删除序号为2词条的第2个回答
        修改词条 [替换文字] ?[旧词条文字] ?[序号]：修改词条问题
            示例:
                修改词条 谁是萝莉 谁是萝莉啊？ : 将词条 谁是萝莉 修改为 谁是萝莉啊？
                修改词条 谁是萝莉 --id 2     : 将序号为2的词条修改为 谁是萝莉
        查看词条 ?[问题] ?[序号]：查看全部词条或对应词条回答
            示例:
                查看词条:
                    (在群组中使用时): 查看当前群组词条和全局词条
                    (在私聊中使用时): 查看当前私聊词条和全局词条
                查看词条 谁是萝莉   : 查看词条 谁是萝莉 的全部回答
                查看词条 --id 2    : 查看词条序号为2的全部回答
                查看词条 谁是萝莉 --all: 查看全局词条 谁是萝莉 的全部回答
                查看词条 --id 2 --all: 查看全局词条序号为2的全部回答
        查看词条:
            (在群组中使用时): 查看当前群组词条和全局词条
            (在私聊中使用时): 查看当前私聊词条和全局词条
            查看词条 谁是萝莉   : 查看词条 谁是萝莉 的全部回答
            查看词条 --id 2    : 查看词条序号为2的全部回答
            查看词条 谁是萝莉 --all: 查看全局词条 谁是萝莉 的全部回答
            查看词条 --id 2 --all: 查看全局词条序号为2的全部回答

    超级用户额外指令
        在私聊中超级用户额外设置
        指令：
            (全局|私聊)?添加词条\s*?(模糊|正则|图片)?问\s*?(\S*\s?\S*)\s*?答\s?(\S*)：添加问答词条，可重复添加相同问题的不同回答
            全局添加词条
            私聊添加词条
            （私聊情况下）删除词条: 删除私聊词条
            （私聊情况下）修改词条: 修改私聊词条
            通过添加参数 --all才指定全局词条
            示例:
                删除词条 --id 2 --all: 删除全局词条中序号为2的词条
            用法与普通用法相同
    """,
    type="application",
    homepage="https://github.com/HibiKier/nonebot-plugin-zxpm",
    config=Config,
    supported_adapters=inherit_supported_adapters(
        "nonebot_plugin_alconna",
        "nonebot_plugin_uninfo",
    ),
    extra={"author": "HibiKier"},
)

from .message_handle import *  # noqa: E402, F403
from .read_word_bank import *  # noqa: E402, F403
from .word_handle import *  # noqa: E402, F403



from io import BytesIO

import imagehash
from PIL import Image
from nonebot import logger
from nonebot.typing import T_State
from nonebot.adapters import Bot, Event
from nonebot_plugin_uninfo import Uninfo
from zhenxun_utils.http_utils import AsyncHttpx
from nonebot_plugin_alconna import At, Text, UniMsg

from ._model import WordBank
from ._data_source import get_img_and_at_list


async def check(
    bot: Bot,
    event: Event,
    message: UniMsg,
    session: Uninfo,
    state: T_State,
) -> bool:
    text = message.extract_plain_text().strip()
    img_list, at_list = get_img_and_at_list(message)
    problem = text
    if not text and len(img_list) == 1:
        try:
            r = await AsyncHttpx.get(img_list[0])
            problem = str(imagehash.average_hash(Image.open(BytesIO(r.content))))
        except Exception as e:
            logger.warning(f"词条获取图片失败 {type(e)}:{e}...")
    if at_list:
        temp = ""
        # TODO: 支持更多消息类型
        for msg in message:
            if isinstance(msg, At):
                temp += f"[at:{msg.target}]"
            elif isinstance(msg, Text):
                temp += msg.text
        problem = temp
    if event.is_tome() and bot.config.nickname:
        if isinstance(message[0], At) and message[0].target == bot.self_id:
            problem = f"[at:{bot.self_id}]{problem}"
        elif problem and bot.config.nickname:
            nickname = [nk for nk in bot.config.nickname if str(message).startswith(nk)]
            problem = nickname[0] + problem if nickname else problem
    group_id = session.group.id if session.group else None
    if problem and (await WordBank.check_problem(group_id, problem) is not None):
        state["problem"] = problem  # type: ignore
        return True
    return False



from nonebot_plugin_alconna import At, Image, UniMessage, UniMsg
from nonebot_plugin_alconna import At as alcAt
from nonebot_plugin_alconna import Image as alcImage
from nonebot_plugin_alconna import Text as alcText

from zhenxun_utils.image_utils import ImageTemplate
from zhenxun_utils.message import MessageUtils

from ._config import ScopeType
from ._model import WordBank


def get_img_and_at_list(message: UniMsg) -> tuple[list[str], list[str]]:
    """获取图片和at数据

    参数:
        message: UniMsg

    返回:
        tuple[list[str], list[str]]: 图片列表，at列表
    """
    img_list, at_list = [], []
    for msg in message:
        if isinstance(msg, alcImage):
            img_list.append(msg.url)
        elif isinstance(msg, alcAt):
            at_list.append(msg.target)
    return img_list, at_list


def get_problem(message: UniMsg) -> str:
    """获取问题内容

    参数:
        message: UniMsg

    返回:
        str: 问题文本
    """
    problem = ""
    a, b = True, True
    for msg in message:
        if isinstance(msg, alcText | str):
            msg = str(msg)
            if "问" in msg and a:
                a = False
                split_text = msg.split("问")
                if len(split_text) > 1:
                    problem += "问".join(split_text[1:])
            if b:
                if "答" in problem:
                    b = False
                    problem = problem.split("答")[0]
                elif "答" in msg:
                    b = False
                    # problem += "答".join(msg.split("答")[:-1])
                    problem += msg.split("答")[0]
        if not a and not b:
            break
        if isinstance(msg, alcAt):
            problem += f"[at:{msg.target}]"
    return problem


def get_answer(message: UniMsg) -> UniMessage:
    """获取at时回答

    参数:
        message: UniMsg

    返回:
        str: 回答内容
    """
    temp_message = None
    answer = ""
    index = 0
    for msg in message:
        index += 1
        if isinstance(msg, alcText | str):
            msg = str(msg)
            if "答" in msg:
                answer += "答".join(msg.split("答")[1:])
                break
    # if answer:
    temp_message = message[index:]
    temp_message.insert(0, alcText(answer))
    return temp_message


class WordBankManage:
    @classmethod
    async def update_word(
        cls,
        replace: str,
        problem: str = "",
        index: int | None = None,
        group_id: str | None = None,
        word_scope: ScopeType = ScopeType.GROUP,
    ) -> tuple[str, str]:
        """修改群词条

        参数:
            params: 参数
            group_id: 群号
            word_scope: 词条范围

        返回:
            tuple[str, str]: 处理消息，替换的旧词条
        """
        return await cls.__word_handle(
            problem, group_id, "update", index, None, word_scope, replace
        )

    @classmethod
    async def delete_word(
        cls,
        problem: str,
        index: int | None = None,
        aid: int | None = None,
        group_id: str | None = None,
        word_scope: ScopeType = ScopeType.GROUP,
    ) -> tuple[str, str]:
        """删除群词条

        参数:
            params: 参数
            index: 指定下标
            aid: 指定回答下标
            group_id: 群号
            word_scope: 词条范围

        返回:
            tuple[str, str]: 处理消息，空
        """
        return await cls.__word_handle(
            problem, group_id, "delete", index, aid, word_scope
        )

    @classmethod
    async def __word_handle(
        cls,
        problem: str,
        group_id: str | None,
        handle_type: str,
        index: int | None = None,
        aid: int | None = None,
        word_scope: ScopeType = ScopeType.GLOBAL,
        replace_problem: str = "",
    ) -> tuple[str, str]:
        """词条操作

        参数:
            problem: 参数
            group_id: 群号
            handle_type: 类型
            index: 指定回答下标
            aid: 指定回答下标
            word_scope: 词条范围
            replace_problem: 替换问题内容

        返回:
            tuple[str, str]: 处理消息，替换的旧词条
        """
        if word_scope == ScopeType.GLOBAL:
            group_id = None
        if index is not None:
            problem, code = await cls.__get_problem_str(index, group_id, word_scope)
            if code != 200:
                return problem, ""
        if handle_type == "delete":
            if index:
                problem, _problem_list = await WordBank.get_problem_all_answer(
                    problem, None, group_id, word_scope
                )
                if not _problem_list:
                    return problem, ""
            return (
                ("删除词条成功!", "")
                if await WordBank.delete_group_problem(
                    problem, group_id, aid, word_scope
                )
                else ("词条不存在", "")
            )
        elif handle_type == "update":
            old_problem = await WordBank.update_group_problem(
                problem, replace_problem, group_id, word_scope=word_scope
            )
            return f"修改词条成功!\n{old_problem} -> {replace_problem}", old_problem
        return "类型错误", ""

    @classmethod
    async def __get_problem_str(
        cls,
        idx: int,
        group_id: str | None = None,
        word_scope: ScopeType = ScopeType.GROUP,
    ) -> tuple[str, int]:
        """通过id获取问题字符串

        参数:
            idx: 下标
            group_id: 群号
            word_scope: 获取类型
        """
        if word_scope in [ScopeType.GLOBAL, ScopeType.PRIVATE]:
            all_problem = await WordBank.get_problem_by_scope(word_scope)
        elif group_id:
            all_problem = await WordBank.get_group_all_problem(group_id)
        else:
            raise Exception("词条类型与群组id不能为空")
        if idx < 0 or idx >= len(all_problem):
            return "问题下标id必须在范围内", 999
        return all_problem[idx][0], 200

    @classmethod
    async def show_word(
        cls,
        problem: str | None,
        index: int | None = None,
        group_id: str | None = None,
        word_scope: ScopeType | None = ScopeType.GROUP,
    ) -> UniMessage:
        """获取群词条

        参数:
            problem: 问题
            group_id: 群组id
            word_scope: 词条范围
            index: 指定回答下标
        """
        if problem or index is not None:
            msg_list = []
            problem, _problem_list = await WordBank.get_problem_all_answer(
                problem,  # type: ignore
                index,
                group_id if group_id is not None else None,
                word_scope,
            )
            if not _problem_list:
                return MessageUtils.build_message(problem)
            for msg in _problem_list:
                _text = str(msg)
                if isinstance(msg, At):
                    _text = f"[at:{msg.target}]"
                elif isinstance(msg, Image):
                    _text = msg.url or msg.path
                elif isinstance(msg, list):
                    _text = []
                    for m in msg:
                        __text = str(m)
                        if isinstance(m, At):
                            __text = f"[at:{m.target}]"
                        elif isinstance(m, Image):
                            # TODO: 显示词条回答图片
                            # __text = (m.data["image"], 30, 30)
                            __text = "[图片]"
                        _text.append(__text)
                    _text = "".join(_text)
                msg_list.append(_text)
            column_name = ["序号", "回答内容"]
            data_list = []
            for index, msg in enumerate(msg_list):
                data_list.append([index, msg])
            template_image = await ImageTemplate.table_page(
                f"词条 {problem} 的回答", None, column_name, data_list
            )
            return MessageUtils.build_message(template_image)
        else:
            result = []
            if group_id:
                _problem_list = await WordBank.get_group_all_problem(group_id)
            elif word_scope is not None:
                _problem_list = await WordBank.get_problem_by_scope(word_scope)
            else:
                raise Exception("群组id和词条范围不能都为空")
            global_problem_list = await WordBank.get_problem_by_scope(ScopeType.GLOBAL)
            if not _problem_list and not global_problem_list:
                return MessageUtils.build_message("未收录任何词条...")
            column_name = ["序号", "关键词", "匹配类型", "收录用户"]
            data_list = [list(s) for s in _problem_list]
            for i in range(len(data_list)):
                data_list[i].insert(0, i)
            group_image = await ImageTemplate.table_page(
                "群组内词条" if group_id else "私聊词条", None, column_name, data_list
            )
            result.append(group_image)
            if global_problem_list:
                data_list = [list(s) for s in global_problem_list]
                for i in range(len(data_list)):
                    data_list[i].insert(0, i)
                global_image = await ImageTemplate.table_page(
                    "全局词条", None, column_name, data_list
                )
                result.append(global_image)
            return MessageUtils.build_message(result)



import re
from typing import Any

from nonebot import logger
from nonebot.typing import T_State
from nonebot.params import RegexGroup
from nonebot.adapters import Bot, Event
from nonebot_plugin_alconna import Text
from nonebot_plugin_alconna import Image
from nonebot_plugin_uninfo import Uninfo
from zhenxun_utils.message import MessageUtils
from nonebot.exception import FinishedException
from nonebot.internal.params import Arg, ArgStr
from zhenxun_utils.platform import PlatformUtils
from nonebot_plugin_alconna import Image as alcImage
from nonebot_plugin_alconna.uniseg.tools import reply_fetch
from nonebot_plugin_alconna import Match, Query, UniMsg, Arparma, AlconnaQuery

from ._model import WordBank
from .exception import ImageDownloadError
from ._config import WordType, ScopeType, type2int, scope2int
from ._command import _add_matcher, _del_matcher, _update_matcher
from ._data_source import WordBankManage, get_answer, get_problem, get_img_and_at_list


@_add_matcher.handle()
async def _(
    bot: Bot,
    event: Event,
    session: Uninfo,
    state: T_State,
    message: UniMsg,
    reg_group: tuple[Any, ...] = RegexGroup(),
):
    img_list, at_list = get_img_and_at_list(message)
    user_id = session.user.id
    group_id = session.group.id if session.group else None
    if not group_id and user_id not in bot.config.superusers:
        await MessageUtils.build_message("权限不足捏...").finish(reply_to=True)
    word_scope, word_type, problem = reg_group
    if not word_scope and not group_id:
        word_scope = "私聊"
    if (
        word_scope
        and word_scope in ["全局", "私聊"]
        and user_id not in bot.config.superusers
    ):
        await MessageUtils.build_message("权限不足，无法添加该范围词条...").finish(
            reply_to=True
        )
    if word_type != "图片":
        state["problem_image"] = "YES"
    if reply := await reply_fetch(event, bot):
        if reply.msg:
            message = message[1:]
            message += Text("答") + MessageUtils.template2alc(reply.msg)  # type: ignore
        logger.debug(f"获取词条消息引用消息: {message}")
    temp_problem = message.copy()
    answer = get_answer(message.copy())
    if (not problem or not problem.strip()) and word_type != "图片":
        await MessageUtils.build_message("词条问题不能为空！").finish(reply_to=True)
    if (not answer or not answer.strip()) and not len(img_list) and not len(at_list):
        await MessageUtils.build_message("词条回答不能为空！").finish(reply_to=True)
    state["word_scope"] = word_scope
    state["word_type"] = word_type
    state["problem"] = get_problem(temp_problem)
    state["answer"] = answer.strip()
    logger.info(
        f"添加词条 范围: {word_scope} 类型: {word_type} 问题: {problem} 回答: {answer}"
    )


@_add_matcher.got("problem_image", prompt="请发送该回答设置的问题图片")
async def _(
    bot: Bot,
    session: Uninfo,
    message: UniMsg,
    word_scope: str | None = ArgStr("word_scope"),
    word_type: str | None = ArgStr("word_type"),
    problem: str | None = ArgStr("problem"),
    answer: Any = Arg("answer"),
):
    user_id = session.user.id
    group_id = session.group.id if session.group else None
    try:
        if word_type == "图片":
            problem = next(m for m in message if isinstance(m, alcImage)).url
        elif word_type == "正则" and problem:
            try:
                re.compile(problem)
            except re.error:
                await MessageUtils.build_message(
                    f"添加词条失败，正则表达式 {problem} 非法！"
                ).finish(reply_to=True)
        nickname = None
        if problem and bot.config.nickname:
            nickname = [nk for nk in bot.config.nickname if problem.startswith(nk)]
        if not problem:
            await MessageUtils.build_message("获取问题失败...").finish(reply_to=True)
        platform = PlatformUtils.get_platform(session)
        await WordBank.add_problem_answer(
            user_id,
            (
                group_id
                if group_id and (not word_scope or word_scope == "私聊")
                else "0"
            ),
            scope2int[word_scope] if word_scope else ScopeType.GROUP,
            type2int[word_type] if word_type else WordType.EXACT,
            problem,
            answer,
            nickname[0] if nickname else None,
            platform,
            session.user.id,
        )
    except ImageDownloadError:
        logger.error(f"添加词条 {problem} 错误，图片资源下载失败...")
        await MessageUtils.build_message(
            f"添加词条 {problem} 错误，图片资源下载失败..."
        ).finish()
    except Exception as e:
        if isinstance(e, FinishedException):
            await _add_matcher.finish()
        logger.error(f"添加词条 {problem} 错误...")
        await MessageUtils.build_message(
            f"添加词条 {problem if word_type != '图片' else '图片'} 发生错误！"
        ).finish(reply_to=True)
    if word_type == "图片":
        result = MessageUtils.build_message(
            ["添加词条 ", Image(url=problem), " 成功！"]
        )
    else:
        result = MessageUtils.build_message(f"添加词条 {problem} 成功！")
    await result.send()
    logger.info(f"添加词条 {problem} 成功！")


@_del_matcher.handle()
async def _(
    bot: Bot,
    session: Uninfo,
    problem: Match[str],
    index: Match[int],
    answer_id: Match[int],
    arparma: Arparma,
    all: Query[bool] = AlconnaQuery("all.value", False),
):
    if not problem.available and not index.available:
        await MessageUtils.build_message(
            "此命令之后需要跟随指定词条或id，通过“显示词条“查看"
        ).finish(reply_to=True)
    word_scope = ScopeType.GROUP if session.group else ScopeType.PRIVATE
    if all.result:
        word_scope = ScopeType.GLOBAL
    if gid := session.group.id if session.group else None:
        result, _ = await WordBankManage.delete_word(
            problem.result,
            index.result if index.available else None,
            answer_id.result if answer_id.available else None,
            gid,
            word_scope,
        )
    else:
        if session.user.id not in bot.config.superusers:
            await MessageUtils.build_message("权限不足捏...").finish(reply_to=True)
        result, _ = await WordBankManage.delete_word(
            problem.result,
            index.result if index.available else None,
            answer_id.result if answer_id.available else None,
            None,
            word_scope,
        )
    await MessageUtils.build_message(result).send(reply_to=True)
    logger.info(f"删除词条: {problem.result}")


@_update_matcher.handle()
async def _(
    bot: Bot,
    session: Uninfo,
    replace: str,
    problem: Match[str],
    index: Match[int],
    arparma: Arparma,
    all: Query[bool] = AlconnaQuery("all.value", False),
):
    if not problem.available and not index.available:
        await MessageUtils.build_message(
            "此命令之后需要跟随指定词条或id，通过“显示词条“查看"
        ).finish(reply_to=True)
    word_scope = ScopeType.GROUP if session.group else ScopeType.PRIVATE
    if all.result:
        word_scope = ScopeType.GLOBAL
    gid = session.group.id if session.group else None
    if session.user.id not in bot.config.superusers:
        if not gid:
            await MessageUtils.build_message("权限不足捏...").finish(reply_to=True)
    result, old_problem = await WordBankManage.update_word(
        replace,
        problem.result if problem.available else "",
        index.result if index.available else None,
        gid,
        word_scope,
    )
    await MessageUtils.build_message(result).send(reply_to=True)
    logger.info(f"更新词条词条: {old_problem} -> {replace}")



from nonebot import logger, on_message
from nonebot.typing import T_State
from nonebot_plugin_uninfo import Uninfo

from ._model import WordBank
from ._rule import check

_matcher = on_message(priority=6, block=True, rule=check)


@_matcher.handle()
async def _(session: Uninfo, state: T_State):
    if problem := state.get("problem"):
        gid = session.group.id if session.group else None
        if result := await WordBank.get_answer(gid, problem):
            await result.send()
            logger.info(f"触发词条 {problem}")



from nonebot import on_regex
from nonebot.permission import SUPERUSER
from nonebot_plugin_uninfo.permission import ADMIN
from nonebot_plugin_alconna import Args, Option, Alconna, on_alconna, store_true

_add_matcher = on_regex(
    r"^(全局|私聊)?添加词条\s*?(模糊|正则|图片)?问(\S*\s?\S*)",
    priority=5,
    block=True,
    permission=ADMIN() | SUPERUSER,
)


_del_matcher = on_alconna(
    Alconna(
        "删除词条",
        Args["problem?", str],
        Option("--all", action=store_true, help_text="所有词条"),
        Option("--id", Args["index", int], help_text="下标id"),
        Option("--aid", Args["answer_id", int], help_text="回答下标id"),
    ),
    priority=5,
    block=True,
    permission=ADMIN() | SUPERUSER,
)


_update_matcher = on_alconna(
    Alconna(
        "修改词条",
        Args["replace", str]["problem?", str],
        Option("--id", Args["index", int], help_text="词条id"),
        Option("--all", action=store_true, help_text="全局词条"),
    ),
    priority=5,
    block=True,
    permission=ADMIN() | SUPERUSER,
)



import nonebot
from nonebot import logger
from tortoise import Tortoise
from tortoise.connection import connections
from tortoise.models import Model as Model_

from ._config import ZxwbConfig, data_dir

driver = nonebot.get_driver()

MODELS: list[str] = []


class DbConnectError(Exception):
    """
    数据库连接错误
    """

    pass


class Model(Model_):
    """
    自动添加模块

    Args:
        Model_: Model
    """

    def __init_subclass__(cls, **kwargs):
        MODELS.append(cls.__module__)


@driver.on_startup
async def _():
    if not ZxwbConfig.zxwb_db_url:
        db_file = data_dir / "db" / "zxwb.db"
        db_file.parent.mkdir(parents=True, exist_ok=True)
        ZxwbConfig.zxwb_db_url = f"sqlite:{db_file.absolute()}"
        logger.info(f"未配置ZXWB数据库连接，使用默认连接: sqlite:{db_file.absolute()}")
    try:
        await Tortoise.init(
            db_url=ZxwbConfig.zxwb_db_url,
            modules={"models": MODELS},
            timezone="Asia/Shanghai",
        )
        await Tortoise.generate_schemas()
        logger.info("ZXWB数据库加载完成!")
    except Exception as e:
        raise DbConnectError(f"ZXWB数据库连接错误... e:{e}") from e


@driver.on_shutdown
async def disconnect():
    await connections.close_all()



import random
import re
import uuid
from datetime import datetime
from pathlib import Path
from typing import Any

import imagehash
from nonebot import logger
from nonebot_plugin_alconna import At, AtAll, Image, Text, UniMessage
from PIL import Image as PilImage
from tortoise import Tortoise, fields
from tortoise.expressions import Q
from typing_extensions import Self
from zhenxun_utils.http_utils import AsyncHttpx
from zhenxun_utils.message import MessageUtils

from ._config import ScopeType, WordType, data_dir, db_type, int2type
from .db_connect import Model
from .exception import ImageDownloadError


def get_img_hash(image_file: str | Path) -> str:
    """获取图片的hash值

    参数:
        image_file: 图片文件路径

    返回:
        str: 哈希值
    """
    hash_value = ""
    try:
        with open(image_file, "rb") as fp:
            hash_value = imagehash.average_hash(PilImage.open(fp))
    except Exception as e:
        logger.warning("获取图片Hash出错", "禁言检测", e=e)
    return str(hash_value)


class WordBank(Model):
    id = fields.IntField(pk=True, generated=True, auto_increment=True)
    """自增id"""
    user_id = fields.CharField(255)
    """用户id"""
    group_id = fields.CharField(255, null=True)
    """群聊id"""
    word_scope = fields.IntField(default=ScopeType.GLOBAL.value)
    """生效范围 0: 全局 1: 群聊 2: 私聊"""
    word_type = fields.IntField(default=WordType.EXACT.value)
    """词条类型 0: 完全匹配 1: 模糊 2: 正则 3: 图片"""
    status = fields.BooleanField()
    """词条状态"""
    problem = fields.TextField()
    """问题，为图片时使用图片hash"""
    answer = fields.TextField()
    """回答"""
    placeholder = fields.TextField(null=True)
    """占位符"""
    image_path = fields.TextField(null=True)
    """使用图片作为问题时图片存储的路径"""
    to_me = fields.CharField(255, null=True)
    """昵称开头时存储的昵称"""
    create_time = fields.DatetimeField(auto_now=True)
    """创建时间"""
    update_time = fields.DatetimeField(auto_now_add=True)
    """更新时间"""
    platform = fields.CharField(255, default="qq")
    """平台"""
    author = fields.CharField(255, null=True, default="")
    """收录人"""

    class Meta:  # type: ignore
        table = "word_bank2"
        table_description = "词条数据库"

    @classmethod
    async def exists(  # type: ignore
        cls,
        user_id: str | None,
        group_id: str | None,
        problem: str,
        answer: str | None,
        word_scope: ScopeType | None = None,
        word_type: WordType | None = None,
    ) -> bool:
        """检测问题是否存在

        参数:
            user_id: 用户id
            group_id: 群号
            problem: 问题
            answer: 回答
            word_scope: 词条范围
            word_type: 词条类型
        """
        query = cls.filter(problem=problem)
        if user_id:
            query = query.filter(user_id=user_id)
        if group_id:
            query = query.filter(group_id=group_id)
        if answer:
            query = query.filter(answer=answer)
        if word_type is not None:
            query = query.filter(word_type=word_type.value)
        if word_scope is not None:
            query = query.filter(word_scope=word_scope.value)
        return await query.exists()

    @classmethod
    async def add_problem_answer(
        cls,
        user_id: str,
        group_id: str | None,
        word_scope: ScopeType,
        word_type: WordType,
        problem: str,
        answer: list[str | Text | At | Image],
        to_me_nickname: str | None = None,
        platform: str = "",
        author: str = "",
    ):
        """添加或新增一个问答

        参数:
            user_id: 用户id
            group_id: 群号
            word_scope: 词条范围,
            word_type: 词条类型,
            problem: 问题, 为图片时是URl
            answer: 回答
            to_me_nickname: at真寻名称
            platform: 所属平台
            author: 收录人id
        """
        # 对图片做额外处理
        image_path = None
        if word_type == WordType.IMAGE:
            _uuid = uuid.uuid1()
            _file = data_dir / "problem" / f"{group_id}" / f"{user_id}_{_uuid}.jpg"
            _file.parent.mkdir(exist_ok=True, parents=True)
            if not await AsyncHttpx.download_file(problem, _file):
                raise ImageDownloadError()
            problem = get_img_hash(_file)
            image_path = f"problem/{group_id}/{user_id}_{_uuid}.jpg"
        new_answer, placeholder_list = await cls._answer2format(
            answer,  # type: ignore
            user_id,
            group_id,
        )
        if not await cls.exists(
            user_id, group_id, problem, new_answer, word_scope, word_type
        ):
            await cls.create(
                user_id=user_id,
                group_id=group_id,
                word_scope=word_scope.value,
                word_type=word_type.value,
                status=True,
                problem=str(problem).strip(),
                answer=new_answer,
                image_path=image_path,
                placeholder=",".join(placeholder_list),
                create_time=datetime.now().replace(microsecond=0),
                update_time=datetime.now().replace(microsecond=0),
                to_me=to_me_nickname,
                platform=platform,
                author=author,
            )

    @classmethod
    async def _answer2format(
        cls,
        answer: list[str | Text | At | AtAll | Image],
        user_id: str,
        group_id: str | None,
    ) -> tuple[str, list[Any]]:
        """将特殊字段转化为占位符，图片，at等

        参数:
            answer: 回答内容
            user_id: 用户id
            group_id: 群号

        返回:
            tuple[str, list[Any]]: 替换后的文本回答内容，占位符
        """
        placeholder_list = []
        text = ""
        index = 0
        for seg in answer:
            placeholder = uuid.uuid1()
            if isinstance(seg, str):
                text += seg
            elif isinstance(seg, Text):
                text += seg.text
            elif seg.type == "face":  # TODO: face貌似无用...
                text += f"[face:placeholder_{placeholder}]"
                placeholder_list.append(seg.data["id"])
            elif isinstance(seg, At | AtAll):
                text += f"[at:placeholder_{placeholder}]"
                placeholder_list.append(seg.target if isinstance(seg, At) else "0")
            elif isinstance(seg, Image) and seg.url:
                text += f"[image:placeholder_{placeholder}]"
                index += 1
                _file = (
                    data_dir
                    / "answer"
                    / f"{group_id or user_id}"
                    / f"{user_id}_{placeholder}.jpg"
                )
                _file.parent.mkdir(exist_ok=True, parents=True)
                logger.debug(f"下载词条回答图片url: {seg.url}", "添加词条")
                if not await AsyncHttpx.download_file(seg.url, _file):
                    pass
                placeholder_list.append(
                    f"answer/{group_id or user_id}/{user_id}_{placeholder}.jpg"
                )
        return text, placeholder_list

    @classmethod
    async def _format2answer(
        cls,
        problem: str,
        answer: str,
        user_id: int,
        group_id: int,
        query: Self | None = None,
    ) -> UniMessage:
        """将占位符转换为实际内容

        参数:
            problem: 问题内容
            answer: 回答内容
            user_id: 用户id
            group_id: 群组id
        """
        if not query:
            query = await cls.get_or_none(
                problem=problem,
                user_id=user_id,
                group_id=group_id,
                answer=answer,
            )
        if not answer:
            answer = str(query.answer)  # type: ignore
        if query and query.placeholder:
            type_list = re.findall(r"\[(.*?):placeholder_.*?]", answer)
            answer_split = re.split(r"\[.*?:placeholder_.*?]", answer)
            placeholder_split = query.placeholder.split(",")
            result_list = []
            for index, ans in enumerate(answer_split):
                result_list.append(ans)
                if index < len(type_list):
                    t = type_list[index]
                    p = placeholder_split[index]
                    if t == "at":
                        if p == "0":
                            result_list.append(AtAll())
                        else:
                            result_list.append(At(flag="user", target=p))
                    elif t == "image":
                        result_list.append(data_dir / p)
            return MessageUtils.build_message(result_list)
        return MessageUtils.build_message(answer)

    @classmethod
    async def check_problem(
        cls,
        group_id: str | None,
        problem: str,
        word_scope: ScopeType | None = None,
        word_type: WordType | None = None,
    ) -> Any:
        """检测是否包含该问题并获取所有回答"""

        query = cls
        if group_id:
            if word_scope:
                query = query.filter(word_scope=word_scope.value)
            else:
                query = query.filter(
                    Q(group_id=group_id) | Q(word_scope=WordType.EXACT.value)
                )
        else:
            query = query.filter(
                Q(word_scope=ScopeType.PRIVATE.value)
                | Q(word_scope=ScopeType.GLOBAL.value)
            )
            if word_type:
                query = query.filter(word_scope=word_type.value)

        # 完全匹配
        if data_list := await query.filter(
            Q(Q(word_type=WordType.EXACT.value) | Q(word_type=WordType.IMAGE.value)),
            Q(problem=problem),
        ).all():
            return data_list
        db = Tortoise.get_connection("default")
        # 模糊匹配：处理 POSITION 和 INSTR 的差异
        if "postgres" in db_type:
            sql = (
                query.filter(word_type=WordType.FUZZY.value).sql()
                + " AND POSITION(problem IN $1) > 0"
            )
            params = [problem]
        elif "sqlite" in db_type:
            sql = (
                query.filter(word_type=WordType.FUZZY.value).sql()
                + " AND INSTR(?, problem) > 0"
            )
            params = [problem]
        elif "mysql" in db_type:
            sql = (
                query.filter(word_type=WordType.FUZZY.value).sql()
                + " AND INSTR(%s, problem) > 0"
            )
            params = [problem]
        else:
            raise Exception(f"Unsupported database type: {db_type}")

        data_list = await db.execute_query_dict(sql, params)
        if data_list:
            return [cls(**data) for data in data_list]

        # 正则匹配
        if "postgres" in db_type:
            sql = (
                query.filter(word_type=WordType.REGEX.value, word_scope__not=999).sql()
                + " AND $1 ~ problem"
            )
            params = [problem]
        elif "sqlite" in db_type:
            # SQLite 不支持 REGEXP，使用 LIKE 替代
            sql = (
                query.filter(word_type=WordType.REGEX.value, word_scope__not=999).sql()
                + " AND problem LIKE ?"
            )
            params = [f"%{problem}%"]
        elif "mysql" in db_type:
            sql = (
                query.filter(word_type=WordType.REGEX.value, word_scope__not=999).sql()
                + " AND problem REGEXP ?"
            )
            params = [problem]

        data_list = await db.execute_query_dict(sql, params)
        return [cls(**data) for data in data_list] if data_list else None

    @classmethod
    async def get_answer(
        cls,
        group_id: str | None,
        problem: str,
        word_scope: ScopeType | None = None,
        word_type: WordType | None = None,
    ) -> UniMessage | None:
        """根据问题内容获取随机回答

        参数:
            user_id: 用户id
            group_id: 群组id
            problem: 问题内容
            word_scope: 词条范围
            word_type: 词条类型
        """
        data_list = await cls.check_problem(group_id, problem, word_scope, word_type)
        if data_list:
            random_answer = random.choice(data_list)
            if random_answer.word_type == WordType.REGEX:
                r = re.search(random_answer.problem, problem)
                has_placeholder = re.search(r"\$(\d)", random_answer.answer)
                if r and r.groups() and has_placeholder:
                    pats = re.sub(r"\$(\d)", r"\\\1", random_answer.answer)
                    random_answer.answer = re.sub(random_answer.problem, pats, problem)
            return (
                await cls._format2answer(
                    random_answer.problem,
                    random_answer.answer,
                    random_answer.user_id,
                    random_answer.group_id,
                    random_answer,
                )
                if random_answer.placeholder
                else MessageUtils.build_message(random_answer.answer)
            )

    @classmethod
    async def get_problem_all_answer(
        cls,
        problem: str,
        index: int | None = None,
        group_id: str | None = None,
        word_scope: ScopeType | None = ScopeType.GLOBAL,
    ) -> tuple[str, list[UniMessage]]:
        """获取指定问题所有回答

        参数:
            problem: 问题
            index: 下标
            group_id: 群号
            word_scope: 词条范围

        返回:
            tuple[str, list[UniMessage]]: 问题和所有回答
        """
        if index is not None:
            # TODO: group_by和order_by不能同时使用
            if group_id and word_scope != ScopeType.GLOBAL:
                _problem = (
                    await cls.filter(group_id=group_id).order_by("create_time")
                    # .group_by("problem")
                    .values_list("problem", flat=True)
                )
            else:
                _problem = (
                    await cls.filter(
                        word_scope=(word_scope or ScopeType.GLOBAL).value
                    ).order_by("create_time")
                    # .group_by("problem")
                    .values_list("problem", flat=True)
                )
            # if index is None and problem not in _problem:
            #     return "词条不存在...", []
            sort_problem = []
            for p in _problem:
                if p not in sort_problem:
                    sort_problem.append(p)
            if index > len(sort_problem) - 1:
                return "下标错误，必须小于问题数量...", []
            problem = sort_problem[index]  # type: ignore
        f = cls.filter(
            problem=problem, word_scope=(word_scope or ScopeType.GLOBAL).value
        )
        if group_id and word_scope != ScopeType.GLOBAL:
            f = f.filter(group_id=group_id)
        answer_list = await f.all()
        if not answer_list:
            return "词条不存在...", []
        return problem, [await cls._format2answer("", "", 0, 0, a) for a in answer_list]

    @classmethod
    async def delete_group_problem(
        cls,
        problem: str,
        group_id: str | None,
        index: int | None = None,
        word_scope: ScopeType = ScopeType.GROUP,
    ):
        """删除指定问题全部或指定回答

        参数:
            problem: 问题文本
            group_id: 群号
            index: 回答下标
            word_scope: 词条范围
        """
        if await cls.exists(None, group_id, problem, None, word_scope):
            if index is not None:
                if group_id:
                    query = await cls.filter(
                        group_id=group_id, problem=problem, word_scope=word_scope.value
                    ).all()
                else:
                    query = await cls.filter(
                        word_scope=word_scope.value, problem=problem
                    ).all()
                await query[index].delete()
            else:
                if group_id:
                    await WordBank.filter(
                        group_id=group_id, problem=problem, word_scope=word_scope.value
                    ).delete()
                else:
                    await WordBank.filter(
                        word_scope=word_scope.value, problem=problem
                    ).delete()
            return True
        return False

    @classmethod
    async def update_group_problem(
        cls,
        problem: str,
        replace_str: str,
        group_id: str | None,
        index: int | None = None,
        word_scope: ScopeType = ScopeType.GROUP,
    ) -> str:
        """修改词条问题

        参数:
            problem: 问题
            replace_str: 替换问题
            group_id: 群号
            index: 问题下标
            word_scope: 词条范围

        返回:
            str: 修改前的问题
        """
        if index is not None:
            if group_id:
                query = await cls.filter(group_id=group_id, problem=problem).all()
            else:
                query = await cls.filter(
                    word_scope=word_scope.value, problem=problem
                ).all()
            tmp = query[index].problem
            query[index].problem = replace_str
            await query[index].save(update_fields=["problem"])
            return tmp
        else:
            if group_id:
                await cls.filter(group_id=group_id, problem=problem).update(
                    problem=replace_str
                )
            else:
                await cls.filter(word_scope=word_scope.value, problem=problem).update(
                    problem=replace_str
                )
            return problem

    @classmethod
    async def get_group_all_problem(cls, group_id: str) -> list[tuple[Any | str]]:
        """获取群聊所有词条

        参数:
            group_id: 群号
        """
        return cls._handle_problem(
            await cls.filter(group_id=group_id).order_by("create_time").all()  # type: ignore
        )

    @classmethod
    async def get_problem_by_scope(cls, word_scope: ScopeType):
        """通过词条范围获取词条

        参数:
            word_scope: 词条范围
        """
        return cls._handle_problem(
            await cls.filter(word_scope=word_scope.value).order_by("create_time").all()  # type: ignore
        )

    @classmethod
    async def get_problem_by_type(cls, word_type: int):
        """通过词条类型获取词条

        参数:
            word_type: 词条类型
        """
        return cls._handle_problem(
            await cls.filter(word_type=word_type).order_by("create_time").all()  # type: ignore
        )

    @classmethod
    def __type2int(cls, value: int) -> str:
        for key, member in WordType.__members__.items():
            if member.value == value:
                return key
        return ""

    @classmethod
    def _handle_problem(cls, problem_list: list["WordBank"]):
        """格式化处理问题

        参数:
            msg_list: 消息列表
        """
        _tmp = []
        result_list = []
        for q in problem_list:
            if q.problem not in _tmp:
                word_type = cls.__type2int(q.word_type)
                # TODO: 获取收录人名称
                problem = (
                    (path / q.image_path, 30, 30) if q.image_path else q.problem,
                    int2type[word_type],
                    # q.author,
                    "-",
                )
                result_list.append(problem)
                _tmp.append(q.problem)
        return result_list

    @classmethod
    async def _move(
        cls,
        user_id: str,
        group_id: str | None,
        problem: str,
        answer: str,
        placeholder: str,
    ):
        """旧词条图片移动方法

        参数:
            user_id: 用户id
            group_id: 群号
            problem: 问题
            answer: 回答
            placeholder: 占位符
        """
        word_scope = ScopeType.GLOBAL
        word_type = WordType.EXACT
        # 对图片做额外处理
        if not await cls.exists(
            user_id, group_id, problem, answer, word_scope, word_type
        ):
            await cls.create(
                user_id=user_id,
                group_id=group_id,
                word_scope=word_scope.value,
                word_type=word_type.value,
                status=True,
                problem=problem,
                answer=answer,
                image_path=None,
                placeholder=placeholder,
                create_time=datetime.now().replace(microsecond=0),
                update_time=datetime.now().replace(microsecond=0),
            )

    @classmethod
    async def _run_script(cls):
        return [
            "ALTER TABLE word_bank2 ADD to_me varchar(255);",  # 添加 to_me 字段
            (
                "ALTER TABLE word_bank2 ALTER COLUMN create_time TYPE timestamp"
                " with time zone USING create_time::timestamp with time zone;"
            ),
            (
                "ALTER TABLE word_bank2 ALTER COLUMN update_time TYPE timestamp"
                " with time zone USING update_time::timestamp with time zone;"
            ),
            "ALTER TABLE word_bank2 RENAME COLUMN user_qq TO user_id;",
            # 将user_qq改为user_id
            "ALTER TABLE word_bank2 ALTER COLUMN user_id TYPE character varying(255);",
            "ALTER TABLE word_bank2 ALTER COLUMN group_id TYPE character varying(255);",
            "ALTER TABLE word_bank2 ADD platform varchar(255) DEFAULT 'qq';",
            "ALTER TABLE word_bank2 ADD author varchar(255) DEFAULT '';",
        ]
